theory TML
imports "Transitions" "Interface" "Utilities"
begin


datatype PC =
    Begin1
  | Begin2
  | Commit1
  | Commit2
  | Read1
  | Read2
  | Write1
  | Write2
  | Write4
  | Write5

type_synonym Status = "PC Interface.Status"

record State = 
  glb :: nat
  store :: "L \<Rightarrow> V"
  status :: "T \<Rightarrow> Status"
  val :: "T \<Rightarrow> V"
  loc :: "T \<Rightarrow> nat"
  addr :: "T \<Rightarrow> L"
  (* Auxilliary var *)
  writer :: "T option"

definition update_status :: "T \<Rightarrow> Status \<Rightarrow> State \<Rightarrow> State"
  where
  "update_status t st s \<equiv> s\<lparr> status := ((status s) (t := st))\<rparr>"

definition update_val :: "T \<Rightarrow> V \<Rightarrow> State \<Rightarrow> State"
  where
  "update_val t v s \<equiv> s\<lparr> val := ((val s) (t := v))\<rparr>"

definition update_loc :: "T \<Rightarrow> nat \<Rightarrow> State \<Rightarrow> State"
  where
  "update_loc t n s \<equiv> s\<lparr> loc := ((loc s) (t := n))\<rparr>"

definition update_addr :: "T \<Rightarrow> L \<Rightarrow> State \<Rightarrow> State"
  where
  "update_addr t l s \<equiv> s\<lparr> addr := ((addr s) (t := l))\<rparr>"

definition update_store :: "L \<Rightarrow> V \<Rightarrow> State \<Rightarrow> State"
  where
  "update_store l v s \<equiv>
    s\<lparr> store := ((store s) (l := v)) \<rparr>"

definition start :: "State \<Rightarrow> bool"
  where
  "start s \<equiv>
      (\<forall> t. status s t = NotStarted)
    \<and> store s : mem_initial
    \<and> glb s = 0
    \<and> writer s = None"

definition tml_pre :: "State \<Rightarrow> T \<Rightarrow> PC Event \<Rightarrow> bool"
  where
  "tml_pre s t e \<equiv>
      e \<noteq> External Cancel
      \<and>
      (e = External Abort \<longrightarrow> status s t = AbortPending)
      \<and>
      (case e of
          External a \<Rightarrow> ext_enabled (status s t) a
        | Internal a \<Rightarrow> status s t = Pending a)"

definition ext_eff :: "State \<Rightarrow> T \<Rightarrow> Action \<Rightarrow> State"
  where
  "ext_eff s0 t a \<equiv>
     case a of
          BeginInv \<Rightarrow> s0 ;; update_status t (Pending Begin1)
        | BeginResp \<Rightarrow> s0 ;; update_status t Ready
        | CommitInv \<Rightarrow> s0 ;; update_status t (Pending Commit1)
        | CommitResp \<Rightarrow> s0 ;; update_status t Committed
        | Abort \<Rightarrow> s0 ;; update_status t Aborted
        | ReadInv l \<Rightarrow> s0 ;; update_addr t l ;; update_status t (Pending Read1)
        | ReadResp v \<Rightarrow> s0 ;; update_status t Ready
        | WriteInv l v \<Rightarrow>
           s0 ;; update_addr t l ;; update_val t v ;; update_status t (Pending Write1)
        | WriteResp \<Rightarrow> s0 ;; update_status t Ready"

definition int_eff :: "State \<Rightarrow> T \<Rightarrow> PC \<Rightarrow> State"
  where
  "int_eff s0 t a \<equiv>
     case a of
         Begin1 \<Rightarrow> s0 ;; update_loc t (glb s0) ;; update_status t (Pending Begin2)
       | Begin2 \<Rightarrow> if even (loc s0 t)
                   then s0 ;; (update_status t BeginResponding)
                   else s0 ;; (update_status t (Pending Begin1))

       | Commit1 \<Rightarrow> let st = if odd (loc s0 t)
                             then (Pending Commit2)
                             else CommitResponding in
                       s0 ;; (update_status t st)
       | Commit2 \<Rightarrow> (s0 \<lparr> glb := (loc s0 t)+1 \<rparr>
                        \<lparr> writer := None \<rparr>
                        ;; update_status t CommitResponding)

       | Read1 \<Rightarrow> s0 ;; update_val t (store s0 (addr s0 t)) ;; (update_status t (Pending Read2))
       | Read2 \<Rightarrow> let st = (if (glb s0 = loc s0 t)
                            then ReadResponding (val s0 t)
                            else AbortPending) in
                    s0 ;; update_status t st

       | Write1 \<Rightarrow> let pc = if (even (loc s0 t))
                            then Write2
                            else Write5 in
                     s0 ;; update_status t (Pending pc)
       | Write2 \<Rightarrow> if (glb s0 = loc s0 t)
                   then (s0 \<lparr> glb := (loc s0 t)+1 \<rparr>
                            \<lparr> writer := Some t \<rparr>
                            ;; (update_status t (Pending Write4)))
                   else s0 ;; (update_status t AbortPending)
       | Write4 \<Rightarrow> s0 ;; update_loc t ((loc s0 t) + 1) ;; (update_status t (Pending Write5))
       | Write5 \<Rightarrow> s0 ;; (update_store (addr s0 t) (val s0 t)) ;; (update_status t WriteResponding)"

definition tml_eff :: "State \<Rightarrow> T \<Rightarrow> PC Event \<Rightarrow> State"
  where
  "tml_eff s t e \<equiv> case e of Internal i \<Rightarrow> int_eff s t i | External a \<Rightarrow> ext_eff s t a"

definition TML :: "(State, PC) DAut"
  where
  "TML \<equiv> \<lparr> DAut.start = start,
           DAut.pre = tml_pre,
           DAut.eff = tml_eff \<rparr>"

declare TML_def [simp]

definition global_inv :: "State \<Rightarrow> bool"
  where
  "global_inv s \<equiv>
     even (glb s) \<longleftrightarrow> writer s = None"

definition is_writer :: "State \<Rightarrow> T \<Rightarrow> bool"
  where
  "is_writer s t \<equiv> writer s = Some t \<and> loc s t = glb s"

definition is_ready :: "State \<Rightarrow> T \<Rightarrow> bool"
  where
  "is_ready s t \<equiv> 
      loc s t \<le> glb s
    \<and> (odd (loc s t) \<longrightarrow> is_writer s t)"

declare is_writer_def [simp]
declare is_ready_def [simp]
declare if_splits [split]

definition txn_inv :: "PC Event \<Rightarrow> State \<Rightarrow> T \<Rightarrow> bool"
  where
  "txn_inv e s at \<equiv> (* TODO should be t not at *)
   tml_pre s at e \<longrightarrow>
   (case e of
      External BeginInv \<Rightarrow> True
      |
      Internal Begin1 \<Rightarrow> True
      |
      Internal Begin2 \<Rightarrow> loc s at \<le> glb s
      |
      External BeginResp \<Rightarrow> (  loc s at \<le> glb s
                             \<and> even (loc s at))
      |
      Internal Read1 \<Rightarrow> is_ready s at
      |
      Internal Read2 \<Rightarrow> (   loc s at = glb s
                         \<and> (odd (loc s at) \<longrightarrow> is_writer s at))
                         \<or>
                        loc s at < glb s
      |
      External (ReadResp v) \<Rightarrow>   v = val s at
                              \<and> is_ready s at
      |
      Internal Write1 \<Rightarrow> is_ready s at
      |
      Internal Write2 \<Rightarrow> (  loc s at = glb s
                          \<and> even (loc s at))
                         \<or>
                         loc s at < glb s
      |
      Internal Write4 \<Rightarrow> (even (loc s at) \<and> writer s = Some at \<and> glb s = 1 + loc s at)
      |
      Internal Write5 \<Rightarrow> odd(loc s at) \<and> is_writer s at
      |
      External WriteResp \<Rightarrow> odd(loc s at) \<and> is_writer s at
      |
      Internal Commit1 \<Rightarrow> is_ready s at
      |
      Internal Commit2 \<Rightarrow> odd(loc s at) \<and> is_writer s at
      |
      External CommitResp \<Rightarrow> True
      |
      External Abort \<Rightarrow> True
      |
      External Cancel \<Rightarrow> False
      |
      _ \<Rightarrow> is_ready s at)
"



lemmas unfold_updates =
  update_status_def
  update_loc_def update_addr_def update_val_def
  update_store_def

lemmas unfold_tml =
  start_def ext_eff_def int_eff_def tml_eff_def
  tml_pre_def
  unfold_updates

lemmas unfold_all_tml =
  unfold_tml
  ext_enabled_def
  Transitions.pre_def Transitions.eff_def Transitions.start_def
  Utilities.all_utilities

lemmas splits =
  Action.split PC.split

thm Action.split_asm

(* TODO: this should be called exhaust_event or such. Also, is there a
way to generate this automatically, using [OF ...] or such? *)
lemma Event_split:
  "\<lbrakk>b = External BeginInv \<Longrightarrow> P;
    b = Internal Begin1 \<Longrightarrow> P;
    b = Internal Begin2 \<Longrightarrow> P;
    b = External BeginResp \<Longrightarrow> P;
    \<And> l. b = External (ReadInv l) \<Longrightarrow> P;
    b = Internal Read1 \<Longrightarrow> P;
    b = Internal Read2 \<Longrightarrow> P;
    \<And> v. b = External (ReadResp v) \<Longrightarrow> P;
    \<And> l v. b = External (WriteInv l v) \<Longrightarrow> P;
    b = Internal Write1 \<Longrightarrow> P;
    b = Internal Write2 \<Longrightarrow> P;
    b = Internal Write4 \<Longrightarrow> P;
    b = Internal Write5 \<Longrightarrow> P;
    b = External WriteResp \<Longrightarrow> P;
    b = External CommitInv \<Longrightarrow> P;
    b = Internal Commit1 \<Longrightarrow> P;
    b = Internal Commit2 \<Longrightarrow> P;
    b = External CommitResp \<Longrightarrow> P;
    b = External Cancel \<Longrightarrow> P;
    b = External Abort \<Longrightarrow> P \<rbrakk>
   \<Longrightarrow>
    P"
apply(cases rule: Event.exhaust[where y=b])
apply(auto simp add: Action.exhaust Event.exhaust PC.exhaust
           split: Action.split Event.split PC.split)
apply (meson Action.exhaust_sel)
by (meson PC.exhaust)

(* Begin *)

lemma txn_inv_pres_begin_inv_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External BeginInv);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml split: if_splits)

lemma txn_inv_pres_begin_inv_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (External BeginInv);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml split: if_splits)

lemma txn_inv_pres_begin1_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Begin1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_begin1_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Begin1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_begin2_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Begin2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_begin2_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Begin2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_begin_resp_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External BeginResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_begin_resp_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (External BeginResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

(* Read *)

lemma txn_inv_pres_read_inv_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External (ReadInv l));
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_read_inv_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = External (ReadInv l);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_read1_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Read1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_read1_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Read1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_read2_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Read2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_read2_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Read2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_read_resp_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External (ReadResp v));
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_read_resp_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = External (ReadResp l);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

(* Write *)

lemma txn_inv_pres_write_inv_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External (WriteInv l v));
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_write_inv_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = External (WriteInv l v);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_write1_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (simp_all add: global_inv_def txn_inv_def unfold_all_tml, fastforce+)

lemma txn_inv_pres_write1_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Write1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_write2_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (simp_all add: global_inv_def txn_inv_def unfold_all_tml, fastforce+)

lemma txn_inv_pres_write2_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Write2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (simp_all add: global_inv_def txn_inv_def unfold_all_tml, fastforce+)


lemma txn_inv_pres_write4_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write4);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_write4_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Write4);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma txn_inv_pres_write5_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write5);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_write5_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Write5);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_write_resp_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External WriteResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_write_resp_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = External WriteResp;
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

(* Commit *)

lemma txn_inv_pres_commit_inv_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External CommitInv);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_commit_inv_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (External CommitInv);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_commit1_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Commit1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_commit1_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Commit1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_commit2_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Commit2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_commit2_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (Internal Commit2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (simp_all add: global_inv_def txn_inv_def unfold_all_tml, fastforce+)

lemma txn_inv_pres_commit_resp_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External CommitResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_commit_resp_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (External CommitResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_cancel_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External Cancel);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(simp add: tml_pre_def tml_eff_def unfold_updates)

lemma txn_inv_pres_cancel_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (External Cancel);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(simp add: tml_pre_def tml_eff_def unfold_updates)


lemma txn_inv_pres_abort_self:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External Abort);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) at"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma txn_inv_pres_abort_other:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv b s t;
    txn_inv a s at;
    t \<noteq> at;
    a = (External Abort);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   txn_inv b (tml_eff s at a) t"
by(cases rule: Event_split[where b=b],
   simp_all add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

(* global_inv *)

(* Begin *)

lemma global_inv_pres_begin_inv:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External BeginInv);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_begin1:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Begin1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_begin2:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Begin2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_begin_resp:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External BeginResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)

(* Read *)

lemma global_inv_pres_read_inv:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External (ReadInv l));
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_read1:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Read1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_read2:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Read2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_read_resp:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External (ReadResp v));
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)

(* write *)

lemma global_inv_pres_write_inv:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External (WriteInv l v));
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_write1:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_write2:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_write4:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write4);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_write5:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Write5);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_write_resp:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External WriteResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


(* commit *)

lemma global_inv_pres_commit_inv:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External CommitInv);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)


lemma global_inv_pres_commit1:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Commit1);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma global_inv_pres_commit2:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (Internal Commit2);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (auto simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma global_inv_pres_commit_resp:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External CommitResp);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma global_inv_pres_abort:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External Abort);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def tml_eff_def unfold_updates)
  (simp add: global_inv_def txn_inv_def unfold_all_tml)

lemma global_inv_pres_cancel:
  "\<lbrakk>reach TML s;
    global_inv s;
    txn_inv a s at;
    a = (External Cancel);
    tml_pre s at a\<rbrakk>
   \<Longrightarrow>
   global_inv (tml_eff s at a)"
by(simp add: tml_pre_def)

lemma total_inv_initial:
 "start s \<Longrightarrow> global_inv s \<and> (\<forall>e t. txn_inv e s t)"
by(simp add:  global_inv_def txn_inv_def unfold_all_tml
           split: Event.split Action.split PC.split )

lemma global_inv_preserved:
  "\<lbrakk>reach TML s; global_inv s; txn_inv a s at; tml_pre s at a \<rbrakk> \<Longrightarrow> global_inv (tml_eff s at a)"
by (cases rule: Event_split[where b=a])
   (simp_all add: global_inv_pres_cancel global_inv_pres_abort
                  global_inv_pres_begin_inv global_inv_pres_begin1 global_inv_pres_begin2 global_inv_pres_begin_resp
                  global_inv_pres_read_inv global_inv_pres_read1 global_inv_pres_read2 global_inv_pres_read_resp
                  global_inv_pres_write_inv global_inv_pres_write1 global_inv_pres_write2 global_inv_pres_write4 global_inv_pres_write5 global_inv_pres_write_resp
                  global_inv_pres_commit_inv global_inv_pres_commit1 global_inv_pres_commit2 global_inv_pres_commit_resp)

lemma txn_inv_preserved_self:
  "\<lbrakk>reach TML s; global_inv s; txn_inv a s at; tml_pre s at a \<rbrakk> \<Longrightarrow> txn_inv b (tml_eff s at a) at"
by (cases rule: Event_split[where b=a])
   (simp_all add: txn_inv_pres_cancel_self txn_inv_pres_abort_self
                  txn_inv_pres_begin_inv_self txn_inv_pres_begin1_self txn_inv_pres_begin2_self txn_inv_pres_begin_resp_self
                  txn_inv_pres_read_inv_self txn_inv_pres_read1_self txn_inv_pres_read2_self txn_inv_pres_read_resp_self
                  txn_inv_pres_write_inv_self txn_inv_pres_write1_self txn_inv_pres_write2_self txn_inv_pres_write4_self txn_inv_pres_write5_self txn_inv_pres_write_resp_self
                  txn_inv_pres_commit_inv_self txn_inv_pres_commit1_self txn_inv_pres_commit2_self txn_inv_pres_commit_resp_self
                  txn_inv_pres_cancel_other txn_inv_pres_abort_other
                  txn_inv_pres_begin_inv_other txn_inv_pres_begin1_other txn_inv_pres_begin2_other txn_inv_pres_begin_resp_other
                  txn_inv_pres_read_inv_other txn_inv_pres_read1_other txn_inv_pres_read2_other txn_inv_pres_read_resp_other
                  txn_inv_pres_write_inv_other txn_inv_pres_write1_other txn_inv_pres_write2_other txn_inv_pres_write4_other txn_inv_pres_write5_other txn_inv_pres_write_resp_other
                  txn_inv_pres_commit_inv_other txn_inv_pres_commit1_other txn_inv_pres_commit2_other txn_inv_pres_commit_resp_other)


lemma txn_inv_preserved_other:
  "\<lbrakk>reach TML s; global_inv s; txn_inv b s t; txn_inv a s at; t \<noteq> at; tml_pre s at a \<rbrakk> \<Longrightarrow> txn_inv b (tml_eff s at a) t"
by (cases rule: Event_split[where b=a])
   (simp_all add: txn_inv_pres_cancel_other txn_inv_pres_abort_other
                  txn_inv_pres_begin_inv_other txn_inv_pres_begin1_other txn_inv_pres_begin2_other txn_inv_pres_begin_resp_other
                  txn_inv_pres_read_inv_other txn_inv_pres_read1_other txn_inv_pres_read2_other txn_inv_pres_read_resp_other
                  txn_inv_pres_write_inv_other txn_inv_pres_write1_other txn_inv_pres_write2_other txn_inv_pres_write4_other txn_inv_pres_write5_other txn_inv_pres_write_resp_other
                  txn_inv_pres_commit_inv_other txn_inv_pres_commit1_other txn_inv_pres_commit2_other txn_inv_pres_commit_resp_other)


lemma total_inv:
  "invariant (ioa TML) (\<lambda> s. global_inv s \<and> (\<forall> e t. txn_inv e s t))"
apply(rule invariant_intro)
  apply(rule total_inv_initial, simp)
  apply(unfold preserved_def, auto simp add: global_inv_preserved)
  apply(metis TML_def txn_inv_preserved_self txn_inv_preserved_other)
done



end
